/**
 * 
 */
package org.jboss.as.ejb3.metadata.processor;

import org.jboss.logging.Logger;
import org.jboss.metadata.ejb.jboss.JBossEnterpriseBeanMetaData;
import org.jboss.metadata.ejb.jboss.JBossEnterpriseBeansMetaData;
import org.jboss.metadata.ejb.jboss.JBossMessageDrivenBeanMetaData;
import org.jboss.metadata.ejb.jboss.JBossMetaData;
import org.jboss.metadata.process.ProcessingException;
import org.jboss.metadata.process.processor.JBossMetaDataProcessor;

/**
 * Responsible for setting the default JMS resource adapter name to be used for EJB3.x Message Driven Beans which
 *  do *not* explicitly set the resource adapter name
 * 
 * @author Jaikiran Pai
 * @version $Revision: $
 */
public class DefaultJMSResourceAdapterMetadataProcessor implements JBossMetaDataProcessor<JBossMetaData>
{

   private static Logger logger = Logger.getLogger(DefaultJMSResourceAdapterMetadataProcessor.class);

   private String defaultResourceAdapterName;

   /**
    * Constructs a {@link DefaultJMSResourceAdapterMetadataProcessor} 
    *  
    * @param defaultJMSResourceAdapterName The default resource adapter name to be used
    *       for EJB3.x MDBs which do not explicitly set the resource adapter name 
    */
   public DefaultJMSResourceAdapterMetadataProcessor(String defaultJMSResourceAdapterName)
   {
      this.defaultResourceAdapterName = defaultJMSResourceAdapterName;
   }

   /**
    * Sets the default JMS resource adapter name to be used for EJB3.x Message Driven Beans which
    * do *not* explicitly set the resource adapter name
    */
   public JBossMetaData process(JBossMetaData metadata) throws ProcessingException
   {
      // we are only concerned with EJB3.x
      if (!metadata.isEJB3x())
      {
         return metadata;
      }
      JBossEnterpriseBeansMetaData enterpriseBeans = metadata.getEnterpriseBeans();
      if (enterpriseBeans == null)
      {
         return metadata;
      }
      for (JBossEnterpriseBeanMetaData enterpriseBean : enterpriseBeans)
      {
         // only check for MDBs
         if (!enterpriseBean.isMessageDriven())
         {
            continue;
         }
         JBossMessageDrivenBeanMetaData messageDrivenBean = (JBossMessageDrivenBeanMetaData) enterpriseBean;
         String raName = messageDrivenBean.getResourceAdapterName();
         // if ra name is not explicitly specified then set the default ra name
         if (raName == null || raName.trim().isEmpty())
         {
            messageDrivenBean.setResourceAdapterName(this.defaultResourceAdapterName);
            if (logger.isTraceEnabled())
            {
               logger.trace("Set the default resource adapter name to " + this.defaultResourceAdapterName
                     + " for message driven bean " + messageDrivenBean.getEjbName());
            }
         }
      }
      return metadata;
   }

}
